"""
This module defines the main visitor algorithm transforming ufl expressions
to pymbolic and loopy.
"""
from dune.perftool.error import PerftoolUFLError
from dune.perftool.generation import get_global_context_value, domain
from dune.perftool.ufl.flatoperators import get_operands
from dune.perftool.ufl.modified_terminals import (ModifiedTerminalTracker,
                                                  Restriction,
                                                  )
from dune.perftool.tools import maybe_wrap_subscript
from dune.perftool.options import get_option
from loopy import Reduction

from pymbolic.primitives import (Call,
                                 Product,
                                 Quotient,
                                 Subscript,
                                 Sum,
                                 Variable,
                                 )

from ufl.algorithms import MultiFunction
from ufl.checks import is_cellwise_constant
from ufl import (VectorElement,
                 MixedElement,
                 TensorElement,
                 TensorProductElement,
                 )
from ufl.classes import (FixedIndex,
                         IndexSum,
                         JacobianDeterminant,
                         )

import pymbolic.primitives as prim
import numpy as np


class UFL2LoopyVisitor(ModifiedTerminalTracker):
    def __init__(self, interface, measure, subdomain_id):
        self.interface = interface
        self.interface.visitor = self
        self.measure = measure
        self.subdomain_id = subdomain_id

        # Call base class constructors
        super(UFL2LoopyVisitor, self).__init__()

    def __call__(self, o, do_predicates=False):
        self.current_info = None
        return self._call(o, do_predicates)

    def accumulate(self, o):
        for info in self.interface.list_accumulation_infos(o, self):
            self.current_info = info
            expr = self._call(o, False)
            if expr != 0:
                from dune.perftool.sympy import simplify_pymbolic_expression
                expr = simplify_pymbolic_expression(expr)
                self.interface.generate_accumulation_instruction(expr, self)

    def _call(self, o, do_predicates):
        # Reset state variables
        self.indexmap = {}
        self.indices = None
        self._indices_backup = []
        self.test_info = None
        self.trial_info = None
        self.inames = ()
        self.do_predicates = do_predicates

        return self.call(o)

    call = MultiFunction.__call__

    #
    # Form argument/coefficients handlers:
    # This is where the actual domain specific work happens
    #

    def argument(self, o):
        self.interface.initialize_function_spaces(o, self)
        # Update the information on where to accumulate this
        info = self.interface.get_accumulation_info(o, self)
        if o.number() == 0:
            if info != self.current_info[0]:
                self.indices = None
                return 0
            else:
                self.test_info = info
        elif o.number() == 1:
            if info != self.current_info[1]:
                self.indices = None
                return 0
            else:
                self.trial_info = info

        # Correct the restriction on boundary integrals
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE
        leaf_element = o.ufl_element()

        # Select the correct leaf element in the case of this being a mixed finite element
        if isinstance(o.ufl_element(), MixedElement):
            index = self.indices[0]
            assert isinstance(index, int)
            self.indices = self.indices[1:]
            if len(self.indices) == 0:
                self.indices = None

            # For the purpose of basis evaluation, we need to take the leaf element
            leaf_element = leaf_element.extract_component(index)[1]

        if self.grad:
            raise PerftoolUFLError("Gradients should have been transformed to reference gradients!!!")

        if self.reference_grad:
            return self.interface.pymbolic_reference_gradient(leaf_element, restriction, o.number())
        else:
            return self.interface.pymbolic_basis(leaf_element, restriction, o.number())

    def coefficient(self, o):
        # Do something different for trial function and coefficients from jacobian apply
        if o.count() == 0 or o.count() == 1:
            # Correct the restriction on boundary integrals
            restriction = self.restriction
            if self.measure == 'exterior_facet':
                restriction = Restriction.NEGATIVE

            self.interface.initialize_function_spaces(o, self)

            index = None
            if isinstance(o.ufl_element(), MixedElement):
                index = self.indices[0]
                assert isinstance(index, int)
                self.indices = self.indices[1:]
                if len(self.indices) == 0:
                    self.indices = None

            if self.grad:
                raise PerftoolUFLError("Gradients should have been transformed to reference gradients!!!")

            if self.reference_grad:
                if o.count() == 0:
                    return self.interface.pymbolic_trialfunction_gradient(o.ufl_element(), restriction, index)
                else:
                    return self.interface.pymbolic_apply_function_gradient(o.ufl_element(), restriction, index)
            else:
                if o.count() == 0:
                    return self.interface.pymbolic_trialfunction(o.ufl_element(), restriction, index)
                else:
                    return self.interface.pymbolic_apply_function(o.ufl_element(), restriction, index)

        # Check if this is a parameter function
        else:
            raise NotImplementedError("Handling non-symbolic parameter functions is currently reevaluated!")
            # We expect all coefficients to be of type Expression!
            assert isinstance(o, Expression)

            # Determine the name of the parameter function
            name = get_global_context_value("data").object_names[id(o)]

            cellwise_constant = is_cellwise_constant(o)

            # Trigger the generation of code for this thing in the parameter class
            if o.on_intersection:
                self.interface.intersection_parameter_function(name, o, cellwise_constant)
            else:
                self.interface.cell_parameter_function(name, o, self.restriction, cellwise_constant)

            # And return a symbol
            return Variable(name)

    #
    # Handlers for all indexing related stuff
    #

    def indexed(self, o):
        # Handle indices first to allow the aggregate handler to mangle in indices
        # This is necessary in some places where the locality of the tree transformation
        # is not given (easiest example: jacobian_inverse handler, complex example:
        # sum factorziation handler for trial function gradient)
        self._indices_backup.append(self.indices)
        self.indices = self.call(o.ufl_operands[1])

        # Handle the aggregate!
        aggr = self.call(o.ufl_operands[0])

        # self.indices being None means they are already handled during recursion!
        if self.indices is None:
            self.indices = self._indices_backup.pop()
            return aggr
        else:
            indices = self.indices
            self.indices = self._indices_backup.pop()
            return maybe_wrap_subscript(aggr, indices)

    def index_sum(self, o):
        # This implementation fully unrolls the given indexed sum.
        # This is done for a variety of reasons:
        # * It eases handling of the given loopy kernel in terms of schedulability
        # * The compiler would unroll these anyway
        # * It allows handling of arbitrarily bad nesting of ComponentTensor and
        #   ListTensor, which otherwise becomes a *nightmare*.
        index = o.ufl_operands[1][0]
        operands = []

        for i in range(o.dimension()):
            self.indexmap[index] = i
            operands.append(self.call(o.ufl_operands[0]))
            del self.indexmap[index]

        from pymbolic import flattened_sum
        return flattened_sum(tuple(operands))

    def _index_or_fixed_index(self, index):
        if isinstance(index, FixedIndex):
            return index._value
        else:
            if index in self.indexmap:
                return self.indexmap[index]
            else:
                return Variable(self.interface.name_index(index))

    def multi_index(self, o):
        return tuple(self._index_or_fixed_index(i) for i in o)

    def index(self, o):
        return self._index_or_fixed_index(o)

    def list_tensor(self, o):
        if all(isinstance(i, int) for i in self.indices):
            index = self.indices[0]
            self.indices = self.indices[1:]
            return self.call(o.ufl_operands[index])
        else:
            return self.interface.pymbolic_list_tensor(o)

    def component_tensor(self, o):
        assert len(self.indices) == len(o.ufl_operands[1])
        # Update the index mapping
        for i, ind in enumerate(o.ufl_operands[1]):
            self.indexmap[ind] = self.indexmap.get(self.indices[i], self.indices[i])

        self.indices = None
        ret = self.call(o.ufl_operands[0])

        for i, ind in enumerate(o.ufl_operands[1]):
            del self.indexmap[ind]

        return ret

    def identity(self, o):
        return self.interface.pymbolic_identity(o)

    #
    # Handlers for arithmetic operators and functions
    # Those handlers would be valid in any code going from UFL to pymbolic
    #

    def product(self, o):
        return prim.flattened_product(tuple(self.call(op) for op in o.ufl_operands))

    def float_value(self, o):
        return o.value()

    def int_value(self, o):
        return o.value()

    def division(self, o):
        return prim.quotient(self.call(o.ufl_operands[0]), self.call(o.ufl_operands[1]))

    def sum(self, o):
        return prim.flattened_sum(tuple(self.call(op) for op in o.ufl_operands))

    def zero(self, o):
        # UFL has Zeroes with shape. We ignore those indices.
        self.indices = None
        return 0.0

    def _evaluate_function(self, python_func, c_func, vals):
        assert isinstance(vals, tuple)
        if all(isinstance(v, (float, int)) for v in vals):
            return python_func(*vals)
        else:
            return prim.Call(prim.Variable(c_func), vals)

    def abs(self, o):
        if isinstance(o.ufl_operands[0], JacobianDeterminant):
            return self.call(o.ufl_operands[0])
        else:
            return self._evaluate_function(abs, "abs", (self.call(o.ufl_operands[0]),))

    def exp(self, o):
        return self._evaluate_function(np.exp, "exp", (self.call(o.ufl_operands[0]),))

    def sqrt(self, o):
        return self._evaluate_function(np.sqrt, "sqrt", (self.call(o.ufl_operands[0]),))

    def sin(self, o):
        return self._evaluate_function(np.sin, "sin", (self.call(o.ufl_operands[0]),))

    def cos(self, o):
        return self._evaluate_function(np.cos, "cos", (self.call(o.ufl_operands[0]),))

    def power(self, o):
        base = self.call(o.ufl_operands[0])
        exponent = self.call(o.ufl_operands[1])
        if isinstance(exponent, int):
            return prim.Product((base,) * exponent)
        return self._evaluate_function(pow, "pow", (base, exponent))

    def _minmax_impl(self, python_func, c_func, children):
        # Build the maximum of those that are constant
        constants = list(filter(lambda i: isinstance(i, (float, int)), children))
        children = set(filter(lambda i: not isinstance(i, (float, int)), children))

        if constants:
            if len(children) == 0:
                return python_func(constants)
            else:
                children.add(python_func(constants))

        # std::max/min takes but two arguments -> construct a chain!
        ret = children.pop()
        while children:
            # NB: There is a pymbolic node Max/Min, which we are intentionally
            # avoiding here, as the C++ nature is so much more like a function!
            ret = prim.Call(prim.Variable(c_func), (ret, children.pop()))

        return ret

    def max_value(self, o):
        return self._minmax_impl(max, "max", tuple(self.call(op) for op in o.ufl_operands))

    def min_value(self, o):
        return self._minmax_impl(min, "min", tuple(self.call(op) for op in o.ufl_operands))

    #
    # Handler for conditionals, use pymbolic base implementation
    #

    def conditional(self, o):
        condition = self.call(o.ufl_operands[0])
        indices = self.indices

        op1 = self.call(o.ufl_operands[1])
        # Restore indexing information for the second branch
        self.indices = indices
        op2 = self.call(o.ufl_operands[2])

        try:
            evaluated = eval(str(condition))
            if evaluated:
                return op1
            else:
                return op2
        except:
            return prim.If(condition, op1, op2)

    def eq(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               "==",
                               right=self.call(o.ufl_operands[1]))

    def ge(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               ">=",
                               right=self.call(o.ufl_operands[1]))

    def gt(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               ">",
                               right=self.call(o.ufl_operands[1]))

    def le(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               "<=",
                               right=self.call(o.ufl_operands[1]))

    def lt(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               "<",
                               right=self.call(o.ufl_operands[1]))

    def ne(self, o):
        return prim.Comparison(self.call(o.ufl_operands[0]),
                               "!=",
                               right=self.call(o.ufl_operands[1]))

    def and_condition(self, o):
        return prim.LogicalAnd((self.call(o.ufl_operands[0]), self.call(o.ufl_operands[1])))

    def or_condition(self, o):
        return prim.LogicalOr((self.call(o.ufl_operands[0]), self.call(o.ufl_operands[1])))

    def not_condition(self, o):
        return prim.LogicalNot(self.call(o.ufl_operands[0]))

    #
    # Handlers for geometric quantities
    #

    def spatial_coordinate(self, o):
        return self.interface.pymbolic_spatial_coordinate()

    def facet_normal(self, o):
        # The normal must be restricted to be well-defined
        assert self.restriction is not Restriction.NONE

        if self.restriction == Restriction.POSITIVE:
            return self.interface.pymbolic_unit_outer_normal()
        if self.restriction == Restriction.NEGATIVE:
            # It is highly unnatural to have this generator function,
            # but I do run into subtle trouble with return -1*outer
            # as the indexing into the normal happens only later.
            # Not investing more time into this cornercase right now.
            return self.interface.pymbolic_unit_inner_normal()

    def quadrature_weight(self, o):
        return self.interface.pymbolic_quadrature_weight()

    def jacobian_determinant(self, o):
        return self.interface.pymbolic_jacobian_determinant()

    def jacobian_inverse(self, o):
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        assert(len(self.indices) == 2)
        i, j = self.indices
        self.indices = None

        # Implement diagonal jacobians for unrolled matrices!
        if get_option("diagonal_transformation_matrix"):
            if isinstance(i, int) and isinstance(j, int) and i != j:
                return 0

        return self.interface.pymbolic_jacobian_inverse_transposed(i, j, restriction)

    def jacobian(self, o):
        raise NotImplementedError("How did you get Jacobian into your form? We only support JacobianInverse right now. Report!")

    def facet_jacobian_determinant(self, o):
        return self.interface.pymbolic_facet_jacobian_determinant()

    def cell_volume(self, o):
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        return prim.Variable(self.interface.name_cell_volume(restriction))

    def facet_area(self, o):
        return prim.Variable(self.interface.name_facet_area())

    #
    # Equality/Hashability of the visitor
    # In order to pass this visitor into (cached) generator functions, it is beneficial
    # to add dummy hashability to this class. A generator will *never* yield two different
    # results for two different visitor instances. Do not use this for any other purpose.
    #

    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return 0

    def copy(self):
        """ Get a copy of this visitor """
        return type(self)(self.interface, self.measure, self.subdomain_id)
